import threading
from threading import Thread
import sys
import os
import time
import signal
from StringIO import StringIO
import logging
import inspect
import re

class DTester(Thread):

    class InvalidStepFunction(Exception):
        "The function step provided is invalid"

    class StepNotCallable(InvalidStepFunction):
        "The function step provided is not callable"
            
    class StepFirstArgNotDTesterNorSelf(InvalidStepFunction):
        "The function step should be either dtester or self"

    class StepFailed(Exception):
        "The previously executed step failed"
    
    """
    A Class which handles a DTest Thread.
    DTest falls into (eventually) communicating
    processes which may be runned concurrently and
    be coordinated using barrier and ok messages.  
    """
    logger = logging.getLogger("DTester")
    sh  = logging.StreamHandler()
    fmt = logging.Formatter(fmt="## [%(threadName)s]|%(name)s::%(levelname)s:: %(message)s")
    sh.setFormatter(fmt)
    logger.addHandler(sh)
       
    def __init__(self,name,command=None,timeout=None):
        super(DTester,self).__init__(name=name)
        self.command          = command
        self.timeout          = timeout
        self.expectDidTimeOut = False
        self.environ          = None
        self.dtestmaster      = None
        self.runSteps         = []
        self.initializeSteps  = []
        self.stdin            = StringIO("")
        self.stdout           = StringIO("")
        self.stderr           = StringIO("")
        self.nb_steps         = 0

    def setCommand(self, command):
        self.command = command

    def setTimeout(self, timeout):
        self.timeout = timeout

    def setEnviron(self, environ):
        self.environ = environ
        
    def setDTestMaster(self, dtestmaster):
        self.dtestmaster = dtestmaster

    def registerForBarrier(self, barrierId):
        self.dtestmaster.registerForBarrier(self,barrierId)
        
    def barrier(self, barrierId, timeout=None):
        self.dtestmaster.barrier(self,barrierId,timeout)

    def ok(self, *args, **kwargs):
        self.dtestmaster.ok(self, *args, **kwargs)

    ## The following are command related API
    ## which should be implemented by specific DTester subclass
    ## we do provide a default "no-ops" behavior for the base class

    
    def runCommand(self, command=None, environ=None):
        return True

    def expectTimedOut(self,msg=None):
        self.logger.warning("Expect TimeOut: %s" % msg)
        self.expectDidTimeOut = True

    def createExpectTimer(self,timeout=10,msg=""):
        t = threading.Timer(timeout,self.expectTimedOut,kwargs={'msg': msg})
        t.setName("Timer-"+self.getName())
        return t
    
    def expectFromCommand(self, pattern, timeout=None):        
        pat       = re.compile(pattern)
        monitored = StringIO("")
        readed    = ""
        self.stdin.close()
        self.expectDidTimeOut = False
        if timeout!=None:
            t = threading.Timer(timeout,self.expectTimedOut)
            t.start()
            
        while (not pat.search(monitored.getvalue())) and (not self.expectDidTimeOut):
            self.stdout.flush()
            readed = self.stdout.read(1)
            self.logger.debug("Readed = %s" % readed)
            monitored.write(readed)
            self.logger.debug("Monitored = %s" % monitored.getvalue())

        if self.expectDidTimeOut:
            return False
        else:
            if timeout != None:
                t.cancel()
            return True
    
    def sendToCommand(self, string):
        return True
    
    def waitCommandTermination(self, timeout):
        return True
    
    def terminateCommand(self):
        return True    

    def addInitializeStep(self, stepmethod, *args, **kwargs):
        if not callable(stepmethod):
            raise DTester.StepNotCallable, "Invalid initialize step:"+str(stepmethod)+"is not a callable object"
        self.initializeSteps.append((stepmethod,args,kwargs))

    def addRunStep(self, stepmethod, *args, **kwargs):
        if isinstance(stepmethod,type("")):
            if hasattr(self,stepmethod):
                stepmethod_str = stepmethod
                stepmethod = getattr(self.__class__,stepmethod)
                self.logger.debug("stepmethod string <%s> resolved to %r" % (stepmethod_str,stepmethod))
            else:
                raise DTester.InvalidStepFunction, "Stepmethod is a unknown attr string <%>" % stepmethod
            
        if not callable(stepmethod):
            raise DTester.StepNotCallable, "Invalid run step:"+str(stepmethod)+"is not a callable object"

        if stepmethod == DTester.barrier:
            self.addInitializeStep(DTester.registerForBarrier,args[0])
        elif stepmethod == DTester.ok:
            self.nb_steps += 1
        else:
            (iargs, ivarargs, ivarkw, idefaults) = inspect.getargspec(stepmethod)
            if iargs[0] != "dtester" and iargs[0] != "self":
                raise DTester.StepFirstArgNotDTesterNorSelf, "args[0]="+iargs[0]
            
        self.logger.debug("addRunStep :: step %s args : %s", str(stepmethod), str(args))
        self.runSteps.append((stepmethod,args,kwargs))        

    def initialize(self):
        for step in self.initializeSteps:
            self.logger.debug("initialize :: step[0]: %s", str(step[0]))
            self.logger.debug("initialize :: step[1:] %s", str(step[1:]))
            step[0](self,*(step[1]),**(step[2]))
        
    def run(self):
        for step in self.runSteps:
            self.logger.debug("run :: step[0]: %s", str(step[0]))
            self.logger.debug("run :: step[1:] %s", str(step[1:]))
            if (len(step[1]) > 0) or (len(step[2]) > 0):
                try:
                    step[0](self,*(step[1]),**(step[2]))
                except NotImplementedError, err:
                    self.logger.error("Step <%r> failed <%r>" % (step[0],err))
                    self.ok(False,desc="Step <%r> failed <%r>" % (step[0],err))
            else:
                try:
                    step[0](self)
                except NotImplementedError, err:
                    self.logger.error("Step <%r> failed <%r>" % (step[0],err))
                    self.ok(False,desc="Step <%r> failed <%r>" % (step[0],err))
        self.logger.info("DTester <%s> has terminated" % self.getName())
